# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Internal Imports
from firststreet.models.geometry import Geometry


class LocationDetail:
    """A Location Detail Object parent

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.fsid = response.get('fsid')


class LocationDetailProperty(LocationDetail):
    """Creates a Location Detail Property object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        super().__init__(response)
        self.fsid = response.get('fsid')
        self.streetNumber = response.get('streetNumber')
        self.route = response.get('route')
        self.city = response.get('city')
        self.zipCode = response.get('zipCode')
        self.zcta = response.get('zcta')
        self.neighborhood = response.get('neighborhood')
        self.tract = response.get('tract')
        self.county = response.get('county')
        self.cd = response.get('cd')
        self.state = response.get('state')
        self.footprintId = response.get('footprintId')
        self.elevation = response.get('elevation')
        self.fema = response.get('fema')
        if response.get('geometry'):
            self.geometry = Geometry(response.get('geometry'))


class LocationDetailNeighborhood(LocationDetail):
    """Creates a Location Detail Neighborhood object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        super().__init__(response)
        self.fsid = response.get('fsid')
        self.city = response.get('city')
        self.subtype = response.get('subtype')
        self.county = response.get('county')
        self.state = response.get('state')
        if response.get('geometry'):
            self.geometry = Geometry(response.get('geometry'))
        self.name = response.get('name')


class LocationDetailCity(LocationDetail):
    """Creates a Location Detail City object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        super().__init__(response)
        self.fsid = response.get('fsid')
        self.lsad = response.get('lsad')
        self.zcta = response.get('zcta')
        self.neighborhood = response.get('neighborhood')
        self.county = response.get('county')
        self.state = response.get('state')
        if response.get('geometry'):
            self.geometry = Geometry(response.get('geometry'))
        self.name = response.get('name')


class LocationDetailZcta(LocationDetail):
    """Creates a Location Detail Zcta object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        super().__init__(response)
        self.fsid = response.get('fsid')
        self.city = response.get('city')
        self.county = response.get('county')
        self.state = response.get('state')
        if response.get('geometry'):
            self.geometry = Geometry(response.get('geometry'))
        self.name = response.get('name')


class LocationDetailTract(LocationDetail):
    """Creates a Location Detail Tract object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        super().__init__(response)
        self.fsid = response.get('fsid')
        self.fips = response.get('fips')
        self.county = response.get('county')
        self.state = response.get('state')
        if response.get('geometry'):
            self.geometry = Geometry(response.get('geometry'))


class LocationDetailCounty(LocationDetail):
    """Creates a Location Detail County object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        super().__init__(response)
        self.fsid = response.get('fsid')
        self.city = response.get('city')
        self.zcta = response.get('zcta')
        self.fips = response.get('fips')
        is_coastal = response.get('isCoastal')
        if is_coastal:
            self.isCoastal = True
        else:
            self.isCoastal = False
        self.cd = response.get('cd')
        self.state = response.get('state')
        if response.get('geometry'):
            self.geometry = Geometry(response.get('geometry'))
        self.district = response.get('district')


class LocationDetailCd(LocationDetail):
    """Creates a Location Detail Congressional District object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        super().__init__(response)
        self.fsid = response.get('fsid')
        self.county = response.get('county')
        self.congress = response.get('congress')
        self.state = response.get('state')
        if response.get('geometry'):
            self.geometry = Geometry(response.get('geometry'))
        self.district = response.get('district')


class LocationDetailState(LocationDetail):
    """Creates a Location Detail State object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        super().__init__(response)
        self.fsid = response.get('fsid')
        self.fips = response.get('fips')
        if response.get('geometry'):
            self.geometry = Geometry(response.get('geometry'))
        self.name = response.get('name')


class LocationSummary:
    """A Location Summary Object parent

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.fsid = response.get('fsid')
        self.riskDirection = response.get('riskDirection')
        self.environmentalRisk = response.get('environmentalRisk')
        self.historic = response.get('historic')
        self.adaptation = response.get('adaptation')


class LocationSummaryProperty(LocationSummary):
    """Creates a Location Detail Property object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        super().__init__(response)
        self.floodFactor = response.get('floodFactor')


class LocationSummaryOther(LocationSummary):
    """Creates a Location Detail Property object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        super().__init__(response)
        self.properties = response.get('properties')



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Standard Imports
import asyncio

# External Imports
import logging
import tqdm
import aiohttp

# Internal Imports
import firststreet.errors as e

DEFAULT_SUMMARY_VERSION = 'v1'


class Http:
    """This class handles the communication with the First Street Foundation API by constructing and sending the HTTP
        requests, and handles any errors during the execution.
        Attributes:
            api_key (str): A string specifying the API key.
            version (str): The version to call the API with
        Methods:
            execute: Sends a request to the First Street Foundation API for the specified endpoint
        """

    def __init__(self, api_key, version=None):
        if version is None:
            version = DEFAULT_SUMMARY_VERSION

        self.api_key = api_key
        self.options = {'url': "https://astg.firststreet.org",
                        'headers': {
                            'Content-Encoding': 'gzip',
                            'Content-Type': 'text/html',
                            'User-Agent': 'python/firststreet',
                            'Accept': 'application/vnd.api+json',
                            'Authorization': 'Bearer %s' % api_key
                        }}
        self.version = version

    async def endpoint_execute(self, endpoints, limit=100):
        """Asynchronously calls each endpoint and returns the JSON responses
        Args:
            endpoints (list): List of endpoints to get
            limit (int): max number of connections to make
        Returns:
            The list of JSON responses corresponding to each endpoint
        """

        connector = aiohttp.TCPConnector(limit_per_host=limit)
        session = aiohttp.ClientSession(connector=connector)

        try:
            tasks = [self.execute(endpoint, session) for endpoint in endpoints]
            ret = [await f
                   for f in tqdm.tqdm(asyncio.as_completed(tasks), total=len(tasks))]

        finally:
            await session.close()

        return ret

    async def execute(self, endpoint, session):
        """Executes the endpoint for the given endpoint with the open session
        Args:
            endpoint (str): The endpoint to get from
            session (ClientSession): The open session
        Returns:
            The JSON reponse or an empty body if error
        Raises:
            _network_error: if an error occurs
        """
        headers = self.options.get('headers')

        retry = 0
        while retry < 5:

            try:
                async with session.get(endpoint[0], headers=headers) as response:

                    rate_limit = self._parse_rate_limit(response.headers)
                    body = await response.json(content_type=None)

                    if response.status != 200 and response.status != 404:
                        raise self._network_error(self.options, body.get('error'), rate_limit)

                    error = body.get("error")
                    if error:
                        fsid = endpoint[1]
                        product = endpoint[2]
                        product_subtype = endpoint[3]

                        if product == 'adaptation' and product_subtype == 'detail':
                            return {'adaptationId': fsid}

                        elif product == 'historic' and product_subtype == 'event':
                            return {'eventId': fsid}

                        else:
                            return {'fsid': fsid}

                    return body

            except asyncio.TimeoutError:
                logging.info("Timeout error for fsid: {} at {}. Retry {}".format(endpoint[1], endpoint[0], retry))
                retry += 1
                await asyncio.sleep(1)

            except aiohttp.ClientError as ex:
                logging.error("{} error while getting fsid: {} from {}".format(ex.__class__, endpoint[1], endpoint[0]))
                return {'fsid': endpoint[1]}

        logging.error("Timeout error after 5 retries for fsid: {} from {}".format(endpoint[1], endpoint[0]))
        return {'fsid': endpoint[1]}

    @staticmethod
    def _parse_rate_limit(headers):
        """Parses the rate limit form the header
        Args:
            headers (CIMultiDictProxy): The header returned from the response
        Returns:
            The rate limit information
        """
        return {'limit': headers.get('x-ratelimit-limit'), 'remaining': headers.get('x-ratelimit-remaining'),
                'reset': headers.get('x-ratelimit-reset'), 'requestId': headers.get('x-request-id')}

    @staticmethod
    def _network_error(options, error, rate_limit):
        """Handles any network errors as a result of the First Street Foundation API
        Args:
            options (dict): The options used in the header of the response
            error (dict): The body returned from the request call
            rate_limit (dict): The rate limit information
        Returns:
            A First Street error class
        """
        status = int(error.get('code'))

        if not status == 429:
            message = "Network Error {}: {}".format(status, error.get('message'))
        else:
            message = "Network Error {}: {}. Limit: {}. Remaining: {}. Reset: {}".format(status, error.get('message'),
                                                                                         rate_limit.get('limit'),
                                                                                         rate_limit.get('remaining'),
                                                                                         rate_limit.get('reset'))

        return {
            401: e.UnauthorizedError(message=message,
                                     attachments={"options": options, "rate_limit": rate_limit}),
            406: e.NotAcceptableError(message=message,
                                      attachments={"options": options, "rate_limit": rate_limit}),
            429: e.RateLimitError(message=message, attachments={"options": options, "rate_limit": rate_limit}),
            500: e.InternalError(message=message, attachments={"options": options, "rate_limit": rate_limit}),
            503: e.OfflineError(message=message, attachments={"options": options, "rate_limit": rate_limit}),
        }.get(status,
              e.UnknownError(message=message, attachments={"options": options, "rate_limit": rate_limit}))



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation


class ProbabilityChance:
    """Creates a Probability Chance object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.fsid = response.get('fsid')
        self.chance = response.get('chance')


class ProbabilityCount:
    """Creates a Probability Count object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.fsid = response.get('fsid')
        self.count = response.get('count')


class ProbabilityCountSummary:
    """Creates a Probability Count Summary object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.fsid = response.get('fsid')
        self.state = response.get('state')
        self.city = response.get('city')
        self.zcta = response.get('zcta')
        self.neighborhood = response.get('neighborhood')
        self.tract = response.get('tract')
        self.county = response.get('county')
        self.cd = response.get('cd')


class ProbabilityCumulative:
    """Creates a Probability Cumulative object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.fsid = response.get('fsid')
        self.cumulative = response.get('cumulative')


class ProbabilityDepth:
    """Creates a Probability Depth object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.fsid = response.get('fsid')
        self.depth = response.get('depth')



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

RATE_LIMIT = 'Request rate limited. Free for non-commercial use for up to 10 requests per minute! To increase ' \
             'your rate limit, please contact api@firststreet.com'
UNAUTHORIZED = 'Unauthorized Access. invalid API key provided.'
UNKNOWN = 'Unknown error, please check your request and try again.'
INTERNAL = 'Internal Server Error.'
NO_BODY = 'No body returned from response.'
NOT_FOUND = 'The specified object could not be found.'
OFFLINE = 'API is currently offline, try again later.'
NOT_ACCEPTABLE = 'You requested a format that is\'t JSON.'
NETWORK_ERROR = 'Network error, check host name.'
DEFAULT_ERROR = 'Unknown Client error.'
ENDPOINT_ERROR = 'HTTP Error: No endpoint provided for request.'
INVALID_ARGUMENT = 'Argument provided was invalid.'


class FirstStreetError(Exception):

    def __init__(self, message=DEFAULT_ERROR, attachments=None):
        super().__init__(message)
        self.message = message
        self.attachments = attachments


class RateLimitError(FirstStreetError):

    def __init__(self, message=RATE_LIMIT, attachments=None):
        super().__init__(message, attachments)


class UnauthorizedError(FirstStreetError):

    def __init__(self, message=UNAUTHORIZED, attachments=None):
        super().__init__(message, attachments)


class UnknownError(FirstStreetError):

    def __init__(self, message=UNKNOWN, attachments=None):
        super().__init__(message, attachments)


class InternalError(FirstStreetError):

    def __init__(self, message=INTERNAL, attachments=None):
        super().__init__(message, attachments)


class NoBodyError(FirstStreetError):

    def __init__(self, message=NO_BODY, attachments=None):
        super().__init__(message, attachments)


class NotFoundError(FirstStreetError):

    def __init__(self, message=NOT_FOUND, attachments=None):
        super().__init__(message, attachments)


class OfflineError(FirstStreetError):

    def __init__(self, message=OFFLINE, attachments=None):
        super().__init__(message, attachments)


class NotAcceptableError(FirstStreetError):

    def __init__(self, message=NOT_ACCEPTABLE, attachments=None):
        super().__init__(message, attachments)


class NetworkError(FirstStreetError):

    def __init__(self, message=NETWORK_ERROR, attachments=None):
        super().__init__(message, attachments)


class EndpointError(FirstStreetError):

    def __init__(self, message=ENDPOINT_ERROR, attachments=None):
        super().__init__(message, attachments)


class InvalidArgument(Exception):

    def __init__(self, message):
        super().__init__(message)
        self.message = INVALID_ARGUMENT + " Provided Arg: %s" % message


class MissingAPIKeyError(Exception):
    pass



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Standard Imports
import datetime
import logging
import os

# External Imports
import pandas as pd


def to_csv(data, product, product_subtype, location_type=None, output_dir=None):
    """Receives a list of data, a product, a product subtype, and a location to create a CSV

    Args:
        data (list): A list of FSF object
        product (str): The overall product to call
        product_subtype (str): The product subtype (if suitable)
        location_type (str): The location lookup type (if suitable)
        output_dir (str): The output directory to save the generated csvs
    """

    date = datetime.datetime.today().strftime('%Y_%m_%d_%H_%M_%S')

    # Set file name to the current date, time, and product
    if location_type:
        file_name = "_".join([date, product, product_subtype, location_type]) + ".csv"
    else:
        file_name = "_".join([date, product, product_subtype]) + ".csv"

    if not output_dir:
        output_dir = os.getcwd() + "/data_csv"

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Format the data for each product
    if product == 'adaptation':

        if product_subtype == 'detail':
            df = format_adaptation_detail(data)

        elif product_subtype == 'summary':
            df = format_adaptation_summary(data)

        else:
            raise NotImplementedError

    elif product == 'probability':
        if product_subtype == 'chance':
            df = format_probability_chance(data)

        elif product_subtype == 'count':
            df = format_probability_count(data)

        elif product_subtype == 'count-summary':
            df = format_probability_count_summary(data)

        elif product_subtype == 'cumulative':
            df = format_probability_cumulative(data)

        elif product_subtype == 'depth':
            df = format_probability_depth(data)
        else:
            raise NotImplementedError

    elif product == 'environmental':
        if product_subtype == 'precipitation':
            df = format_environmental_precipitation(data)

        else:
            raise NotImplementedError

    elif product == 'historic':
        if product_subtype == 'event':
            df = format_historic_event(data)

        elif product_subtype == 'summary':
            if location_type == 'property':
                df = format_historic_summary_property(data)

            else:
                df = format_historic_summary(data)

        else:
            raise NotImplementedError

    elif product == 'location':
        if product_subtype == 'detail':
            if location_type == 'property':
                df = format_location_detail_property(data)

            elif location_type == 'neighborhood':
                df = format_location_detail_neighborhood(data)

            elif location_type == 'city':
                df = format_location_detail_city(data)

            elif location_type == 'zcta':
                df = format_location_detail_zcta(data)

            elif location_type == 'tract':
                df = format_location_detail_tract(data)

            elif location_type == 'county':
                df = format_location_detail_county(data)

            elif location_type == 'cd':
                df = format_location_detail_cd(data)

            elif location_type == 'state':
                df = format_location_detail_state(data)

            else:
                raise NotImplementedError

        elif product_subtype == 'summary':

            if location_type == 'property':
                df = format_location_summary_property(data)

            else:
                df = format_location_summary(data)

        else:
            raise NotImplementedError

    elif product == 'fema':
        if product_subtype == 'nfip':
            df = format_fema_nfip(data)
        else:
            raise NotImplementedError

    else:
        raise NotImplementedError

    # Export CSVs
    df.fillna(pd.NA, inplace=True)
    df.to_csv(output_dir + '/' + file_name, index=False)
    logging.info("CSV generated to '{}'.".format(output_dir + '/' + file_name))


def format_adaptation_detail(data):
    """Reformat the list of data to Adaptation Detail format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('type').explode('scenario').reset_index(drop=True)
    df.drop(["serving", "geometry"], axis=1, inplace=True)
    df['adaptationId'] = df['adaptationId'].astype('Int64').apply(str)
    df['returnPeriod'] = df['returnPeriod'].astype('Int64').apply(str)
    return df[['adaptationId', 'name', 'type', 'scenario', 'conveyance', 'returnPeriod']]


def format_adaptation_summary(data):
    """Reformat the list of data to Adaptation Summary format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('adaptation').reset_index(drop=True)
    df['fsid'] = df['fsid'].apply(str)
    return df[['fsid', 'adaptation', 'properties']]


def format_probability_chance(data):
    """Reformat the list of data to Probability Chance format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('chance').reset_index(drop=True)
    if not df['chance'].isna().values.all():
        df = pd.concat([df.drop(['chance'], axis=1), df['chance'].apply(pd.Series)], axis=1)
        df = df.explode('data').reset_index(drop=True)
        df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
        df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
        df['fsid'] = df['fsid'].astype('Int64').apply(str)
        df['year'] = df['year'].astype('Int64').apply(str)
        df['threshold'] = df['threshold'].astype('Int64').apply(str)
        df['low'] = df['low'].round(3)
        df['mid'] = df['mid'].round(3)
        df['high'] = df['high'].round(3)
    else:
        df['fsid'] = df['fsid'].apply(str)
        df.drop(['chance'], axis=1, inplace=True)
        df['year'] = pd.NA
        df['threshold'] = pd.NA
        df['low'] = pd.NA
        df['mid'] = pd.NA
        df['high'] = pd.NA

    return df[['fsid', 'year', 'threshold', 'low', 'mid', 'high']]


def format_probability_count(data):
    """Reformat the list of data to Probability Count format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('count').reset_index(drop=True)
    if not df['count'].isna().values.all():
        df = pd.concat([df.drop(['count'], axis=1), df['count'].apply(pd.Series)], axis=1)
        df = df.explode('data').reset_index(drop=True)
        df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
        df = df.explode('data').reset_index(drop=True)
        df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
        df = pd.concat([df.drop(['count'], axis=1), df['count'].apply(pd.Series)], axis=1)
        df['fsid'] = df['fsid'].astype('Int64').apply(str)
        df['year'] = df['year'].astype('Int64').apply(str)
        df['returnPeriod'] = df['returnPeriod'].astype('Int64').apply(str)
        df['bin'] = df['bin'].astype('Int64').apply(str)
        df['low'] = df['low'].astype('Int64').apply(str)
        df['mid'] = df['mid'].astype('Int64').apply(str)
        df['high'] = df['high'].astype('Int64').apply(str)
    else:
        df['fsid'] = df['fsid'].apply(str)
        df.drop(['count'], axis=1, inplace=True)
        df['year'] = pd.NA
        df['returnPeriod'] = pd.NA
        df['bin'] = pd.NA
        df['low'] = pd.NA
        df['mid'] = pd.NA
        df['high'] = pd.NA

    return df[['fsid', 'year', 'returnPeriod', 'bin', 'low', 'mid', 'high']]


def format_probability_count_summary(data):
    """Reformat the list of data to Probability Count-Summary format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    # listed_data = [{'fsid': attr.fsid, 'location': [{'state': attr.state}, {'city': attr.city}, {'zcta': attr.zcta},
    #                                                 {'neighborhood': attr.neighborhood}, {'tract': attr.tract},
    #                                                 {'county': attr.county}, {'cd': attr.cd}]} for attr in data]
    listed_data = [{'fsid': attr.fsid, 'location': [{'location': 'state', 'data': attr.state},
                                                    {'location': 'city', 'data': attr.city},
                                                    {'location': 'zcta', 'data': attr.zcta},
                                                    {'location': 'neighborhood', 'data': attr.neighborhood},
                                                    {'location': 'tract', 'data': attr.tract},
                                                    {'location': 'county', 'data': attr.county},
                                                    {'location': 'cd', 'data': attr.cd}]} for attr in data]

    df = pd.DataFrame(listed_data).explode('location').reset_index(drop=True)
    df.rename(columns={'fsid': 'fsid_placeholder'}, inplace=True)
    df = pd.concat([df.drop(['location'], axis=1), df['location'].apply(pd.Series)], axis=1)
    df = df.explode('data').reset_index(drop=True)
    df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
    df.rename(columns={'fsid': 'location_fips', 'name': 'location_name'}, inplace=True)

    if 'count' in df:
        df = df.explode('count').reset_index(drop=True)
        df = pd.concat([df.drop(['count'], axis=1), df['count'].apply(pd.Series)], axis=1)
        df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid_placeholder': 'fsid'}, inplace=True)
        if 'subtype' not in df:
            df['subtype'] = pd.NA
        df['fsid'] = df['fsid'].astype('Int64').apply(str)
        df['location_fips'] = df['location_fips'].astype('Int64').apply(str)
        df['year'] = df['year'].astype('Int64').apply(str)
        df['low'] = df['low'].astype('Int64').apply(str)
        df['mid'] = df['mid'].astype('Int64').apply(str)
        df['high'] = df['high'].astype('Int64').apply(str)

    else:
        df['location_fips'] = pd.NA
        df['location_name'] = pd.NA
        df['year'] = pd.NA
        df['low'] = pd.NA
        df['mid'] = pd.NA
        df['high'] = pd.NA

    return df[['fsid', 'location', 'location_fips', 'location_name', 'subtype', 'year', 'low', 'mid', 'high']]


def format_probability_cumulative(data):
    """Reformat the list of data to Probability Cumulative format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('cumulative').reset_index(drop=True)
    if not df['cumulative'].isna().values.all():
        df = pd.concat([df.drop(['cumulative'], axis=1), df['cumulative'].apply(pd.Series)], axis=1)
        df = df.explode('data').reset_index(drop=True)
        df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
        df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
        df['fsid'] = df['fsid'].astype('Int64').apply(str)
        df['year'] = df['year'].astype('Int64').apply(str)
        df['threshold'] = df['threshold'].astype('Int64').apply(str)
        df['low'] = df['low'].round(3)
        df['mid'] = df['mid'].round(3)
        df['high'] = df['high'].round(3)
    else:
        df['fsid'] = df['fsid'].apply(str)
        df.drop(['cumulative'], axis=1, inplace=True)
        df['year'] = pd.NA
        df['threshold'] = pd.NA
        df['low'] = pd.NA
        df['mid'] = pd.NA
        df['high'] = pd.NA

    return df[['fsid', 'year', 'threshold', 'low', 'mid', 'high']]


def format_probability_depth(data):
    """Reformat the list of data to Probability Depth format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('depth').reset_index(drop=True)
    if not df['depth'].isna().values.all():
        df = pd.concat([df.drop(['depth'], axis=1), df['depth'].apply(pd.Series)], axis=1)
        df = df.explode('data').reset_index(drop=True)
        df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
        df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
        df['fsid'] = df['fsid'].astype('Int64').apply(str)
        df['year'] = df['year'].astype('Int64').apply(str)
        df['returnPeriod'] = df['returnPeriod'].astype('Int64').apply(str)
        df['low'] = df['low'].astype('Int64').apply(str)
        df['mid'] = df['mid'].astype('Int64').apply(str)
        df['high'] = df['high'].astype('Int64').apply(str)
    else:
        df['fsid'] = df['fsid'].astype('Int64').apply(str)
        df.drop(['depth'], axis=1, inplace=True)
        df['year'] = pd.NA
        df['returnPeriod'] = pd.NA
        df['low'] = pd.NA
        df['mid'] = pd.NA
        df['high'] = pd.NA

    return df[['fsid', 'year', 'returnPeriod', 'low', 'mid', 'high']]


def format_environmental_precipitation(data):
    """Reformat the list of data to Environmental Precipitation format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('projected').reset_index(drop=True)
    if not df['projected'].isna().values.all():
        df = pd.concat([df.drop(['projected'], axis=1), df['projected'].apply(pd.Series)], axis=1)
        df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
        df['fsid'] = df['fsid'].astype('Int64').apply(str)
        df['year'] = df['year'].astype('Int64').apply(str)
        df['low'] = df['low'].round(3)
        df['mid'] = df['mid'].round(3)
        df['high'] = df['high'].round(3)
    else:
        df['fsid'] = df['fsid'].apply(str)
        df.drop(['projected'], axis=1, inplace=True)
        df['year'] = pd.NA
        df['low'] = pd.NA
        df['mid'] = pd.NA
        df['high'] = pd.NA

    return df[['fsid', 'year', 'low', 'mid', 'high']]


def format_historic_event(data):
    """Reformat the list of data to Historic Event format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data])
    df.drop(["geometry"], axis=1, inplace=True)
    if not df['properties'].isna().values.all():
        df = pd.concat([df.drop(['properties'], axis=1), df['properties'].apply(pd.Series)], axis=1)
        df['eventId'] = df['eventId'].astype('Int64').apply(str)
        df['month'] = df['month'].astype('Int64').apply(str)
        df['year'] = df['year'].astype('Int64').apply(str)
        df['returnPeriod'] = df['returnPeriod'].astype('Int64').apply(str)
        df['total'] = df['total'].astype('Int64').apply(str)
        df['affected'] = df['affected'].astype('Int64').apply(str)
        df.rename(columns={'total': 'propertiesTotal', 'affected': 'propertiesAffected'}, inplace=True)
    else:
        df['eventId'] = df['eventId'].astype('Int64').apply(str)
        df.drop(['properties'], axis=1, inplace=True)
        df['propertiesTotal'] = pd.NA
        df['propertiesAffected'] = pd.NA

    return df[['eventId', 'name', 'month', 'year', 'returnPeriod', 'type', 'propertiesTotal', 'propertiesAffected']]


def format_historic_summary_property(data):
    """Reformat the list of data to Historic Summary format for property

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('historic').reset_index(drop=True)
    if not df['historic'].isna().values.all():
        df = pd.concat([df.drop(['historic'], axis=1), df['historic'].apply(pd.Series)], axis=1)
        df['fsid'] = df['fsid'].astype('Int64').apply(str)
        df['eventId'] = df['eventId'].astype('Int64').apply(str)
        df['depth'] = df['depth'].astype('Int64').apply(str)
    else:
        df['fsid'] = df['fsid'].astype('Int64').apply(str)
        df.drop(['historic'], axis=1, inplace=True)
        df['eventId'] = pd.NA
        df['name'] = pd.NA
        df['type'] = pd.NA
        df['depth'] = pd.NA

    return df[['fsid', 'eventId', 'name', 'type', 'depth']]


def format_historic_summary(data):
    """Reformat the list of data to Historic Summary format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('historic').reset_index(drop=True)
    if not df['historic'].isna().values.all():
        df = pd.concat([df.drop(['historic'], axis=1), df['historic'].apply(pd.Series)], axis=1)
        df['fsid'] = df['fsid'].astype('Int64').apply(str)
        df = df.explode('data').reset_index(drop=True)
        df = pd.concat([df.drop(['data'], axis=1), df['data'].apply(pd.Series)], axis=1)
        df['eventId'] = df['eventId'].astype('Int64').apply(str)
        df['bin'] = df['bin'].astype('Int64').apply(str)
        df['count'] = df['count'].astype('Int64').apply(str)
    else:
        df['fsid'] = df['fsid'].astype('Int64').apply(str)
        df.drop(['historic'], axis=1, inplace=True)
        df['eventId'] = pd.NA
        df['name'] = pd.NA
        df['type'] = pd.NA
        df['bin'] = pd.NA
        df['count'] = pd.NA

    return df[['fsid', 'eventId', 'name', 'type', 'bin', 'count']]


def format_location_detail_property(data):
    """Reformat the list of data to Location Detail format for property

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('neighborhood').reset_index(drop=True)
    df.rename(columns={'fsid': 'fsid_placeholder'}, inplace=True)

    if not df['city'].isna().values.all():
        df = pd.concat([df.drop(['city'], axis=1), df['city'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'city_fips', 'name': 'city_name'}, inplace=True)
    else:
        df.drop(['city'], axis=1, inplace=True)
        df['city_fips'] = pd.NA
        df['city_name'] = pd.NA

    if not df['neighborhood'].isna().values.all():
        df = pd.concat([df.drop(['neighborhood'], axis=1), df['neighborhood'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'neighborhood_fips', 'name': 'neighborhood_name'}, inplace=True)
    else:
        df.drop(['neighborhood'], axis=1, inplace=True)
        df['neighborhood_fips'] = pd.NA
        df['neighborhood_name'] = pd.NA

    if not df['tract'].isna().values.all():
        df = pd.concat([df.drop(['tract'], axis=1), df['tract'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'tract_fips', 'name': 'tract_name'}, inplace=True)
    else:
        df.drop(['tract'], axis=1, inplace=True)
        df['tract_fips'] = pd.NA
        df['tract_name'] = pd.NA

    if not df['county'].isna().values.all():
        df = pd.concat([df.drop(['county'], axis=1), df['county'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'county_fips', 'name': 'county_name'}, inplace=True)
    else:
        df.drop(['county'], axis=1, inplace=True)
        df['county_fips'] = pd.NA
        df['county_name'] = pd.NA

    if not df['cd'].isna().values.all():
        df = pd.concat([df.drop(['cd'], axis=1), df['cd'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'cd_fips', 'name': 'cd_name'}, inplace=True)
    else:
        df.drop(['cd'], axis=1, inplace=True)
        df['cd_fips'] = pd.NA
        df['cd_name'] = pd.NA

    if not df['state'].isna().values.all():
        df = pd.concat([df.drop(['state'], axis=1), df['state'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'state_fips', 'name': 'state_name'}, inplace=True)
    else:
        df.drop(['state'], axis=1, inplace=True)
        df['state_fips'] = pd.NA
        df['state_name'] = pd.NA

    df.rename(columns={'fsid_placeholder': 'fsid'}, inplace=True)
    df['fsid'] = df['fsid'].apply(str)
    df['city_fips'] = df['city_fips'].astype('Int64').apply(str)
    df['neighborhood_fips'] = df['neighborhood_fips'].astype('Int64').apply(str)
    df['tract_fips'] = df['tract_fips'].astype('Int64').apply(str)
    df['county_fips'] = df['county_fips'].astype('Int64').apply(str)
    df['cd_fips'] = df['cd_fips'].astype('Int64').apply(str)
    df['state_fips'] = df['state_fips'].astype('Int64').apply(str).apply(lambda x: x.zfill(2))
    df['footprintId'] = df['footprintId'].astype('Int64').apply(str)
    df['elevation'] = df['city_fips'].apply(str)
    return df[['fsid', 'streetNumber', 'route', 'city_fips', 'city_name', 'zipCode', 'neighborhood_fips',
               'neighborhood_name', 'tract_fips', 'county_fips', 'county_name', 'cd_fips',
               'cd_name', 'state_fips', 'state_name', 'footprintId', 'elevation', 'fema']]


def format_location_detail_neighborhood(data):
    """Reformat the list of data to Location Detail format for neighborhood

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('city').explode('county').reset_index(drop=True)
    df.rename(columns={'fsid': 'fsid_placeholder', 'name': 'name_placeholder'}, inplace=True)

    if not df['city'].isna().values.all():
        df = pd.concat([df.drop(['city'], axis=1), df['city'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'city_fips', 'name': 'city_name'}, inplace=True)
    else:
        df.drop(['city'], axis=1, inplace=True)
        df['city_fips'] = pd.NA
        df['city_name'] = pd.NA

    if not df['county'].isna().values.all():
        df = pd.concat([df.drop(['county'], axis=1), df['county'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'county_fips', 'name': 'county_name'}, inplace=True)
    else:
        df.drop(['county'], axis=1, inplace=True)
        df['county_fips'] = pd.NA
        df['county_name'] = pd.NA

    if not df['state'].isna().values.all():
        df = pd.concat([df.drop(['state'], axis=1), df['state'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'state_fips', 'name': 'state_name'}, inplace=True)
    else:
        df.drop(['state'], axis=1, inplace=True)
        df['state_fips'] = pd.NA
        df['state_name'] = pd.NA

    df.rename(columns={'fsid_placeholder': 'fsid', 'name_placeholder': 'name'}, inplace=True)
    df['fsid'] = df['fsid'].apply(str)
    df['city_fips'] = df['city_fips'].astype('Int64').apply(str)
    df['county_fips'] = df['county_fips'].astype('Int64').apply(str)
    df['state_fips'] = df['state_fips'].astype('Int64').apply(str).apply(lambda x: x.zfill(2))
    return df[['fsid', 'name', 'city_fips', 'city_name', 'county_fips', 'county_name', 'subtype',
               'state_fips', 'state_name']]


def format_location_detail_city(data):
    """Reformat the list of data to Location Detail format for city

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('zcta').explode('county') \
        .explode('neighborhood').reset_index(drop=True)
    df.rename(columns={'fsid': 'fsid_placeholder', 'name': 'name_placeholder'}, inplace=True)

    if not df['zcta'].isna().values.all():
        df = pd.concat([df.drop(['zcta'], axis=1), df['zcta'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'zipCode', 'name': 'zcta_name'}, inplace=True)
    else:
        df.drop(['zcta'], axis=1, inplace=True)
        df['zipCode'] = pd.NA
        df['zcta_name'] = pd.NA

    if not df['county'].isna().values.all():
        df = pd.concat([df.drop(['county'], axis=1), df['county'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'county_fips', 'name': 'county_name'}, inplace=True)
    else:
        df.drop(['county'], axis=1, inplace=True)
        df['county_fips'] = pd.NA
        df['county_name'] = pd.NA

    if not df['neighborhood'].isna().values.all():
        df = pd.concat([df.drop(['neighborhood'], axis=1), df['neighborhood'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'neighborhood_fips', 'name': 'neighborhood_name'}, inplace=True)
    else:
        df.drop(['neighborhood'], axis=1, inplace=True)
        df['neighborhood_fips'] = pd.NA
        df['neighborhood_name'] = pd.NA

    if not df['state'].isna().values.all():
        df = pd.concat([df.drop(['state'], axis=1), df['state'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'state_fips', 'name': 'state_name'}, inplace=True)
    else:
        df.drop(['state'], axis=1, inplace=True)
        df['state_fips'] = pd.NA
        df['state_name'] = pd.NA

    df.rename(columns={'fsid_placeholder': 'fsid', 'name_placeholder': 'name'}, inplace=True)
    df['fsid'] = df['fsid'].apply(str)
    df['zipCode'] = df['zipCode'].astype('Int64').apply(str)
    df['county_fips'] = df['county_fips'].astype('Int64').apply(str)
    df['neighborhood_fips'] = df['neighborhood_fips'].astype('Int64').apply(str)
    df['state_fips'] = df['state_fips'].astype('Int64').apply(str).apply(lambda x: x.zfill(2))
    return df[['fsid', 'name', 'lsad', 'zipCode', 'neighborhood_fips', 'neighborhood_name',
               'county_fips', 'county_name', 'state_fips', 'state_name']]


def format_location_detail_zcta(data):
    """Reformat the list of data to Location Detail format for zip

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('city').explode('county')
    df.rename(columns={'fsid': 'fsid_placeholder', 'name': 'name_placeholder'}, inplace=True)

    if not df['city'].isna().values.all():
        df = pd.concat([df.drop(['city'], axis=1), df['city'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'city_fips', 'name': 'city_name'}, inplace=True)
    else:
        df.drop(['city'], axis=1, inplace=True)
        df['city_fips'] = pd.NA
        df['city_name'] = pd.NA

    if not df['county'].isna().values.all():
        df = pd.concat([df.drop(['county'], axis=1), df['county'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'county_fips', 'name': 'county_name'}, inplace=True)
    else:
        df.drop(['county'], axis=1, inplace=True)
        df['county_fips'] = pd.NA
        df['county_name'] = pd.NA

    if not df['state'].isna().values.all():
        df = pd.concat([df.drop(['state'], axis=1), df['state'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'state_fips', 'name': 'state_name'}, inplace=True)
    else:
        df.drop(['state'], axis=1, inplace=True)
        df['state_fips'] = pd.NA
        df['state_name'] = pd.NA

    df.rename(columns={'fsid_placeholder': 'fsid', 'name_placeholder': 'name'}, inplace=True)
    df['fsid'] = df['fsid'].apply(str)
    df['city_fips'] = df['city_fips'].astype('Int64').apply(str)
    df['county_fips'] = df['county_fips'].astype('Int64').apply(str)
    df['state_fips'] = df['state_fips'].astype('Int64').apply(str).apply(lambda x: x.zfill(2))
    return df[['fsid', 'name', 'city_fips', 'city_name', 'county_fips', 'county_name', 'state_fips', 'state_name']]


def format_location_detail_tract(data):
    """Reformat the list of data to Location Detail format for tract

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data])
    df.rename(columns={'fsid': 'fsid_placeholder'}, inplace=True)

    if not df['county'].isna().values.all():
        df = pd.concat([df.drop(['county'], axis=1), df['county'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'county_fips', 'name': 'county_name'}, inplace=True)
    else:
        df.drop(['county'], axis=1, inplace=True)
        df['county_fips'] = pd.NA
        df['county_name'] = pd.NA

    if not df['state'].isna().values.all():
        df = pd.concat([df.drop(['state'], axis=1), df['state'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'state_fips', 'name': 'state_name'}, inplace=True)
    else:
        df.drop(['state'], axis=1, inplace=True)
        df['state_fips'] = pd.NA
        df['state_name'] = pd.NA

    df.rename(columns={'fsid_placeholder': 'fsid', 'name_placeholder': 'name'}, inplace=True)
    df['fsid'] = df['fsid'].apply(str)
    df['county_fips'] = df['county_fips'].astype('Int64').apply(str)
    df['state_fips'] = df['state_fips'].astype('Int64').apply(str).apply(lambda x: x.zfill(2))
    return df[['fsid', 'fips', 'county_fips', 'county_name', 'state_fips', 'state_name']]


def format_location_detail_county(data):
    """Reformat the list of data to Location Detail format for county

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('city').explode('zcta') \
        .explode('cd').reset_index(drop=True)
    df.rename(columns={'fsid': 'fsid_placeholder'}, inplace=True)

    if not df['city'].isna().values.all():
        df = pd.concat([df.drop(['city'], axis=1), df['city'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'city_fips', 'name': 'city_name'}, inplace=True)
    else:
        df.drop(['city'], axis=1, inplace=True)
        df['city_fips'] = pd.NA
        df['city_name'] = pd.NA

    if not df['zcta'].isna().values.all():
        df = pd.concat([df.drop(['zcta'], axis=1), df['zcta'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'zipCode', 'name': 'zcta_name'}, inplace=True)
    else:
        df.drop(['zcta'], axis=1, inplace=True)
        df['zipCode'] = pd.NA
        df['zcta_name'] = pd.NA

    if not df['cd'].isna().values.all():
        df = pd.concat([df.drop(['cd'], axis=1), df['cd'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'cd_fips', 'name': 'cd_name'}, inplace=True)
    else:
        df.drop(['cd'], axis=1, inplace=True)
        df['cd_fips'] = pd.NA
        df['cd_name'] = pd.NA

    if not df['state'].isna().values.all():
        df = pd.concat([df.drop(['state'], axis=1), df['state'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'state_fips', 'name': 'state_name'}, inplace=True)
    else:
        df.drop(['state'], axis=1, inplace=True)
        df['state_fips'] = pd.NA
        df['state_name'] = pd.NA

    df.rename(columns={'fsid_placeholder': 'fsid'}, inplace=True)
    df['fsid'] = df['fsid'].apply(str)
    df['city_fips'] = df['city_fips'].astype('Int64').apply(str)
    df['zipCode'] = df['zipCode'].astype('Int64').apply(str)
    df['cd_fips'] = df['cd_fips'].astype('Int64').apply(str)
    df['state_fips'] = df['state_fips'].astype('Int64').apply(str).apply(lambda x: x.zfill(2))
    return df[['fsid', 'city_fips', 'city_name', 'zipCode', 'fips', 'isCoastal', 'cd_fips',
               'cd_name', 'state_fips', 'state_name']]


def format_location_detail_cd(data):
    """Reformat the list of data to Location Detail format for congressional district

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data]).explode('county')
    df.rename(columns={'fsid': 'fsid_placeholder'}, inplace=True)

    if not df['county'].isna().values.all():
        df = pd.concat([df.drop(['county'], axis=1), df['county'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'county_fips', 'name': 'county_name'}, inplace=True)
    else:
        df.drop(['county'], axis=1, inplace=True)
        df['county_fips'] = pd.NA
        df['county_name'] = pd.NA

    if not df['state'].isna().values.all():
        df = pd.concat([df.drop(['state'], axis=1), df['state'].apply(pd.Series)], axis=1)
        df.rename(columns={'fsid': 'state_fips', 'name': 'state_name'}, inplace=True)
    else:
        df.drop(['state'], axis=1, inplace=True)
        df['state_fips'] = pd.NA
        df['state_name'] = pd.NA

    df.rename(columns={'fsid_placeholder': 'fsid'}, inplace=True)
    df['fsid'] = df['fsid'].apply(str)
    df['county_fips'] = df['county_fips'].astype('Int64').apply(str)
    df['state_fips'] = df['state_fips'].astype('Int64').apply(str).apply(lambda x: x.zfill(2))
    return df[['fsid', 'district', 'county_fips', 'county_name', 'state_fips', 'state_name']]


def format_location_detail_state(data):
    """Reformat the list of data to Location Detail format for state

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data])
    df['fsid'] = df['fsid'].apply(str)
    return df[['fsid', 'name', 'fips']]


def format_location_summary_property(data):
    """Reformat the list of data to Location Summary format for property

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data])
    df['fsid'] = df['fsid'].apply(str)
    df['riskDirection'] = df['riskDirection'].astype('Int64').apply(str)
    df['environmentalRisk'] = df['environmentalRisk'].astype('Int64').apply(str)
    df['historic'] = df['historic'].astype('Int64').apply(str)
    df['adaptation'] = df['adaptation'].astype('Int64').apply(str)
    df['floodFactor'] = df['floodFactor'].astype('Int64').apply(str)
    return df[['fsid', 'floodFactor', 'riskDirection', 'environmentalRisk', 'historic', 'adaptation']]


def format_location_summary(data):
    """Reformat the list of data to Location Summary format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data])

    if not df['properties'].isna().values.all():
        df = pd.concat([df.drop(['properties'], axis=1), df['properties'].apply(pd.Series)], axis=1)
        df.rename(columns={'total': 'propertiesTotal', 'atRisk': 'propertiesAtRisk'}, inplace=True)
    else:
        df.drop(['properties'], axis=1, inplace=True)
        df['propertiesTotal'] = pd.NA
        df['propertiesAtRisk'] = pd.NA

    df['fsid'] = df['fsid'].apply(str)
    df['riskDirection'] = df['riskDirection'].astype('Int64').apply(str)
    df['environmentalRisk'] = df['environmentalRisk'].astype('Int64').apply(str)
    df['historic'] = df['historic'].astype('Int64').apply(str)
    df['adaptation'] = df['adaptation'].astype('Int64').apply(str)
    df['propertiesTotal'] = df['propertiesTotal'].astype('Int64').apply(str)
    df['propertiesAtRisk'] = df['propertiesAtRisk'].astype('Int64').apply(str)
    return df[['fsid', 'riskDirection', 'environmentalRisk', 'propertiesTotal', 'propertiesAtRisk', 'historic']]


def format_fema_nfip(data):
    """Reformat the list of data to Fema Nfip format

    Args:
        data (list): A list of FSF object
    Returns:
        A pandas formatted DataFrame
    """
    df = pd.DataFrame([vars(o) for o in data])
    df['fsid'] = df['fsid'].apply(str)
    df['claimCount'] = df['claimCount'].astype('Int64').apply(str)
    df['policyCount'] = df['policyCount'].astype('Int64').apply(str)
    df['buildingPaid'] = df['buildingPaid'].astype('Int64').apply(str)
    df['contentPaid'] = df['contentPaid'].astype('Int64').apply(str)
    df['buildingCoverage'] = df['buildingCoverage'].astype('Int64').apply(str)
    df['contentCoverage'] = df['contentCoverage'].astype('Int64').apply(str)
    df['iccPaid'] = df['iccPaid'].astype('Int64').apply(str)
    return df[['fsid', 'claimCount', 'policyCount', 'buildingPaid', 'contentPaid', 'buildingCoverage',
               'contentCoverage', 'iccPaid']]



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Standard Imports
import argparse
import os
import logging
from distutils.util import strtobool
import sys

# Internal Imports
import firststreet
from firststreet.errors import InvalidArgument
from firststreet.util import read_fsid_file

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Description for my parser")
    parser.add_argument("-p", "--product", help="Example: adaptation_detail", required=True)
    parser.add_argument("-api_key", "--api_key", required=False)
    parser.add_argument("-v", "--version", required=False)
    parser.add_argument("-i", "--fsids", help="Example: 28,29", required=False,)
    parser.add_argument("-l", "--location", help="Example: property", required=False)
    parser.add_argument("-limit", "--limit", help="Example: 100", required=False, default="100")
    parser.add_argument("-log", "--log", help="Example: False", required=False, default="True")
    parser.add_argument("-f", "--file", help="Example: ./sample.txt", required=False)

    argument = parser.parse_args()

    # Merge FSIDs from file and list input
    fsids = []
    if argument.fsids:
        try:
            fsids += list(map(int, argument.fsids.strip().split(",")))
        except ValueError:
            logging.warning("An invalid fsid list was provided. Please check this input list is a comma-separated " 
                            "list of integers: '{}'".format(argument.fsids))

    if argument.file:
        fsids += read_fsid_file(argument.file)

    # Ensure there is at least a product and FSID
    if fsids:

        # Try to get the API key either from env var or the parameter
        if not argument.api_key:
            env_var_name = 'FSF_API_KEY'
            try:
                api_key = os.environ[env_var_name]
            except KeyError:
                logging.error("`{}` is not set as an Environmental Variable".format(env_var_name))
                sys.exit()
        else:
            api_key = argument.api_key

        fs = firststreet.FirstStreet(api_key, version=argument.version, log=bool(strtobool(argument.log)))

        limit = int(argument.limit)

        if argument.product == 'adaptation.get_detail':
            fs.adaptation.get_detail(fsids, csv=True, limit=limit)

        elif argument.product == 'adaptation.get_summary':
            fs.adaptation.get_summary(fsids, argument.location, csv=True, limit=limit)

        elif argument.product == 'probability.get_depth':
            fs.probability.get_depth(fsids, csv=True, limit=limit)

        elif argument.product == 'probability.get_chance':
            fs.probability.get_chance(fsids, csv=True, limit=limit)

        elif argument.product == 'probability.get_count_summary':
            fs.probability.get_count_summary(fsids, csv=True, limit=limit)

        elif argument.product == 'probability.get_cumulative':
            fs.probability.get_cumulative(fsids, csv=True, limit=limit)

        elif argument.product == 'probability.get_count':
            fs.probability.get_count(fsids, argument.location, csv=True, limit=limit)

        elif argument.product == 'historic.get_event':
            fs.historic.get_event(fsids, csv=True, limit=limit)

        elif argument.product == 'historic.get_summary':
            fs.historic.get_summary(fsids, argument.location, csv=True, limit=limit)

        elif argument.product == 'location.get_detail':
            fs.location.get_detail(fsids, argument.location, csv=True, limit=limit)

        elif argument.product == 'location.get_summary':
            fs.location.get_summary(fsids, argument.location, csv=True, limit=limit)

        elif argument.product == 'fema.get_nfip':
            fs.fema.get_nfip(fsids, argument.location, csv=True, limit=limit)

        elif argument.product == 'environmental.get_precipitation':
            fs.environmental.get_precipitation(fsids, csv=True, limit=limit)

        else:
            logging.error("Product not found. Please check that the argument"
                          " provided is correct: {}".format(argument.product))

    else:
        raise InvalidArgument("No fsids were provided from either a fsid list or a file. "
                              "List: '{}', File Name: '{}'".format(argument.fsids, argument.file))



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Standard Imports
import logging

# Internal Imports
from firststreet.api import csv_format
from firststreet.api.api import Api
from firststreet.models.environmental import EnvironmentalPrecipitation


class Environmental(Api):
    """This class receives a list of fsids and handles the creation of a environmental product from the request.

        Methods:
            get_precipitation: Retrieves a list of Environmental Precipitation for the given list of IDs
        """

    def get_precipitation(self, fsids, csv=False, limit=100, output_dir=None):
        """Retrieves environmental precipitation product data from the First Street Foundation API given a list of FSIDs
        and returns a list of Environmental Precipitation objects.

        Args:
            fsids (list/file): A First Street Foundation IDs or a file of First Street Foundation IDs
            csv (bool): To output extracted data to a csv or not
            limit (int): max number of connections to make
            output_dir (str): The output directory to save the generated csvs
        Returns:
            A list of Adaptation Detail
        """

        # Get data from api and create objects
        api_datas = self.call_api(fsids, "environmental", "precipitation", "county", limit=limit)
        product = [EnvironmentalPrecipitation(api_data) for api_data in api_datas]

        if csv:
            csv_format.to_csv(product, "environmental", "precipitation", "county", output_dir=output_dir)

        logging.info("Environmental Precipitation Data Ready.")

        return product



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Internal Imports
from shapely.geometry import shape


class Geometry:
    """Creates a Geometry object given a response

    Args:
        geometry (dict): A dict of geometry
    """

    def __init__(self, geometry):

        if geometry:
            self.polygon = shape(geometry.get('polygon'))
            self.center = shape(geometry.get('center'))
            if geometry.get('bbox'):
                self.bbox = shape(geometry.get('bbox'))
            else:
                self.bbox = None

    def __eq__(self, other):
        if not isinstance(other, Geometry):
            return NotImplemented

        return self.polygon == other.polygon and self.center == other.center and self.bbox == other.bbox



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation


class EnvironmentalPrecipitation:
    """Creates an Environmental Precipitation object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.fsid = response.get("fsid")
        self.projected = response.get("projected")



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Internal Imports
from firststreet.models.geometry import Geometry


class AdaptationDetail:
    """Creates an Adaptation Detail object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.adaptationId = response.get('adaptationId')
        self.name = response.get('name')
        self.type = response.get('type')
        self.scenario = response.get('scenario')
        self.conveyance = response.get('conveyance')
        self.returnPeriod = response.get('returnPeriod')
        self.serving = response.get('serving')
        self.geometry = Geometry(response.get('geometry'))


class AdaptationSummary:
    """Creates an Adaptation Summary object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.fsid = response.get('fsid')
        self.adaptation = response.get('adaptation')
        self.properties = response.get('properties')



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Standard Imports
import logging

# Internal Imports
from firststreet.api import csv_format
from firststreet.api.api import Api
from firststreet.errors import InvalidArgument
from firststreet.models.historic import HistoricEvent, HistoricSummary


class Historic(Api):
    """This class receives a list of fsids and handles the creation of a historic product from the request.

        Methods:
            get_event: Retrieves a list of Historic Event for the given list of IDs
            get_summary: Retrieves a list of Historic Summary for the given list of IDs
        """

    def get_event(self, fsids, csv=False, limit=100, output_dir=None):
        """Retrieves historic event product data from the First Street Foundation API given a list of FSIDs and
        returns a list of Historic Event objects.

        Args:
            fsids (list/file): A First Street Foundation IDs or a file of First Street Foundation IDs
            csv (bool): To output extracted data to a csv or not
            limit (int): max number of connections to make
            output_dir (str): The output directory to save the generated csvs
        Returns:
            A list of Historic Event
        """

        # Get data from api and create objects
        api_datas = self.call_api(fsids, "historic", "event", None, limit)
        product = [HistoricEvent(api_data) for api_data in api_datas]

        if csv:
            csv_format.to_csv(product, "historic", "event", output_dir=output_dir)

        logging.info("Historic Event Data Ready.")

        return product

    def get_summary(self, fsids, location_type, csv=False, limit=100, output_dir=None):
        """Retrieves historic summary product data from the First Street Foundation API given a list of FSIDs and
        returns a list of Historic Summary objects.

        Args:
            fsids (list/file): A First Street Foundation IDs or a file of First Street Foundation IDs
            location_type (str): The location lookup type
            csv (bool): To output extracted data to a csv or not
            limit (int): max number of connections to make
            output_dir (str): The output directory to save the generated csvs
        Returns:
            A list of Historic Summary
        """

        if not location_type:
            raise InvalidArgument(location_type)
        elif not isinstance(location_type, str):
            raise TypeError("location is not a string")

        # Get data from api and create objects
        api_datas = self.call_api(fsids, "historic", "summary", location_type, limit)
        product = [HistoricSummary(api_data) for api_data in api_datas]

        if csv:
            csv_format.to_csv(product, "historic", "summary", location_type, output_dir=output_dir)

        logging.info("Historic Summary Data Ready.")

        return product



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Internal Imports
from firststreet.models.geometry import Geometry


class HistoricEvent:
    """Creates a Historic Event object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.eventId = response.get('eventId')
        self.name = response.get('name')
        self.month = response.get('month')
        self.year = response.get('year')
        self.returnPeriod = response.get('returnPeriod')
        self.type = response.get('type')
        self.properties = response.get('properties')
        self.geometry = Geometry(response.get('geometry'))


class HistoricSummary:
    """Creates a Historic Summary object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.fsid = response.get('fsid')
        self.historic = response.get('historic')



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Standard Imports
import asyncio

# Internal Imports
import os

from firststreet.errors import InvalidArgument
from firststreet.util import read_fsid_file


class Api:
    """This class handles the calls to the API through the http class

        Attributes:
            http (Http): A http class to connect to the First Street Foundation API
        Methods:
            call_api: Creates an endpoint
        """

    def __init__(self, http):
        self._http = http

    def call_api(self, fsids, product, product_subtype, location, limit=100):
        """Receives FSIDs, a product, a product subtype, and a location to create and call an endpoint to the First
        Street Foundation API.

        Args:
            fsids (list/file): A First Street Foundation IDs or a file of First Street Foundation IDs
            product (str): The overall product to call
            product_subtype (str): The product subtype (if suitable)
            location (str/None): The location lookup type (if suitable)
            limit (int): max number of connections to make
        Returns:
            A list of JSON responses
        """
        if not isinstance(fsids, list):
            if isinstance(fsids, str):
                if os.path.isfile(os.getcwd() + "console/terminal" + fsids):
                    fsids = read_fsid_file(os.getcwd() + "/" + fsids)
                else:
                    raise InvalidArgument("File provided is not a valid file. "
                                          "Please check the file name. '{}'".format(os.path.curdir + str(fsids)))
            else:
                raise InvalidArgument("File provided is not a list or a valid file. "
                                      "Please check the file name. '{}'".format(os.path.curdir + str(fsids)))

        if not fsids:
            raise InvalidArgument(fsids)
        elif not isinstance(fsids, list):
            raise TypeError("location is not a string")

        base_url = self._http.options.get('url')
        version = self._http.version

        # Create the endpoint
        if location:
            endpoints = [("/".join([base_url, version, product, product_subtype, location, str(fsid)]), fsid,
                          product, product_subtype) for fsid in fsids]
        else:
            endpoints = [("/".join([base_url, version, product, product_subtype, str(fsid)]), fsid,
                          product, product_subtype) for fsid in fsids]

        # Asynchronously call the API for each endpoint
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(self._http.endpoint_execute(endpoints, limit))

        return response



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Standard Imports
import logging

# Internal Imports
from firststreet.api import csv_format
from firststreet.api.api import Api
from firststreet.errors import InvalidArgument
from firststreet.models.probability import ProbabilityChance, ProbabilityCount, ProbabilityCountSummary, \
    ProbabilityCumulative, ProbabilityDepth


class Probability(Api):
    """This class receives a list of fsids and handles the creation of a probability product from the request.

        Methods:
            get_depth: Retrieves a list of Probability Depth for the given list of IDs
            get_chance: Retrieves a list of Probability Depth for the given list of IDs
            get_count: Retrieves a list of Probability Depth for the given list of IDs
            get_count_summary: Retrieves a list of Probability Depth for the given list of IDs
            get_cumulative: Retrieves a list of Probability Depth for the given list of IDs
        """

    def get_chance(self, fsids, csv=False, limit=100, output_dir=None):
        """Retrieves probability chance product data from the First Street Foundation API given a list of FSIDs and
        returns a list of Probability Chance objects.

        Args:
            fsids (list/file): A First Street Foundation IDs or a file of First Street Foundation IDs
            csv (bool): To output extracted data to a csv or not
            limit (int): max number of connections to make
            output_dir (str): The output directory to save the generated csvs
        Returns:
            A list of Probability Chance
        """

        # Get data from api and create objects
        api_datas = self.call_api(fsids, "probability", "chance", "property", limit=limit)
        product = [ProbabilityChance(api_data) for api_data in api_datas]

        if csv:
            csv_format.to_csv(product, "probability", "chance", output_dir=output_dir)

        logging.info("Probability Chance Data Ready.")

        return product

    def get_count(self, fsids, location_type, csv=False, limit=100, output_dir=None):
        """Retrieves probability count product data from the First Street Foundation API given a list of FSIDs and
        returns a list of Probability Count objects.

        Args:
            fsids (list/file): A First Street Foundation IDs or a file of First Street Foundation IDs
            location_type (str): The location lookup type
            csv (bool): To output extracted data to a csv or not
            limit (int): max number of connections to make
            output_dir (str): The output directory to save the generated csvs
        Returns:
            A list of Probability Count
        Raises:
            InvalidArgument: The location provided is empty
            TypeError: The location provided is not a string
        """

        if not location_type:
            raise InvalidArgument(location_type)
        elif not isinstance(location_type, str):
            raise TypeError("location is not a string")

        # Get data from api and create objects
        api_datas = self.call_api(fsids, "probability", "count", location_type, limit=limit)
        product = [ProbabilityCount(api_data) for api_data in api_datas]

        if csv:
            csv_format.to_csv(product, "probability", "count", location_type, output_dir=output_dir)

        logging.info("Probability Count Data Ready.")

        return product

    def get_count_summary(self, fsids, csv=False, limit=100, output_dir=None):
        """Retrieves probability Count-Summary product data from the First Street Foundation API given a list of FSIDs
        and returns a list of Probability Count-Summary object.

        Args:
            fsids (list/file): A First Street Foundation IDs or a file of First Street Foundation IDs
            csv (bool): To output extracted data to a csv or not
            limit (int): max number of connections to make
            output_dir (str): The output directory to save the generated csvs
        Returns:
            A list of Probability Count-Summary
        """

        # Get data from api and create objects
        api_datas = self.call_api(fsids, "probability", "count-summary", "property", limit)
        product = [ProbabilityCountSummary(api_data) for api_data in api_datas]

        if csv:
            csv_format.to_csv(product, "probability", "count-summary", output_dir=output_dir)

        logging.info("Probability Count-Summary Data Ready.")

        return product

    def get_cumulative(self, fsids, csv=False, limit=100, output_dir=None):
        """Retrieves probability cumulative product data from the First Street Foundation API given a list of FSIDs and
        returns a list of Probability Cumulative object.

        Args:
            fsids (list/file): A First Street Foundation IDs or a file of First Street Foundation IDs
            csv (bool): To output extracted data to a csv or not
            limit (int): max number of connections to make
            output_dir (str): The output directory to save the generated csvs
        Returns:
            A list of Probability Cumulative
        """

        # Get data from api and create objects
        api_datas = self.call_api(fsids, "probability", "cumulative", "property", limit=limit)
        product = [ProbabilityCumulative(api_data) for api_data in api_datas]

        if csv:
            csv_format.to_csv(product, "probability", "cumulative", output_dir=output_dir)

        logging.info("Probability Cumulative Data Ready.")

        return product

    def get_depth(self, fsids, csv=False, limit=100, output_dir=None):
        """Retrieves probability depth product data from the First Street Foundation API given a list of FSIDs and
        returns a list of Probability Depth objects.

        Args:
            fsids (list/file): A First Street Foundation IDs or a file of First Street Foundation IDs
            csv (bool): To output extracted data to a csv or not
            limit (int): max number of connections to make
            output_dir (str): The output directory to save the generated csvs
        Returns:
            A list of Probability Depth
        """

        # Get data from api and create objects
        api_datas = self.call_api(fsids, "probability", "depth", "property", limit=limit)
        product = [ProbabilityDepth(api_data) for api_data in api_datas]

        if csv:
            csv_format.to_csv(product, "probability", "depth", output_dir=output_dir)

        logging.info("Probability Depth Data Ready.")

        return product



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Standard Imports
import logging

# Internal Imports
from firststreet.api import csv_format
from firststreet.api.api import Api
from firststreet.errors import InvalidArgument
from firststreet.models.location import LocationDetailProperty, LocationDetailNeighborhood, LocationDetailCity, \
    LocationDetailZcta, LocationDetailTract, LocationDetailCounty, LocationDetailCd, \
    LocationDetailState, LocationSummaryProperty, LocationSummaryOther


class Location(Api):
    """This class receives a list of fsids and handles the creation of a location product from the request.

        Methods:
            get_detail: Retrieves a list of Location Details for the given list of IDs
            get_summary: Retrieves a list of Location Summary for the given list of IDs
        """

    def get_detail(self, fsids, location_type, csv=False, limit=100, output_dir=None):
        """Retrieves location detail product data from the First Street Foundation API given a list of FSIDs and
        returns a list of Location Detail objects.

        Args:
            fsids (list/file): A First Street Foundation IDs or a file of First Street Foundation IDs
            location_type (str): The location lookup type
            csv (bool): To output extracted data to a csv or not
            limit (int): max number of connections to make
            output_dir (str): The output directory to save the generated csvs
        Returns:
            A list of Location Detail
        Raises:
            InvalidArgument: The location provided is empty
            TypeError: The location provided is not a string
        """

        if not location_type:
            raise InvalidArgument("No loocation type provided: {}".format(location_type))
        elif not isinstance(location_type, str):
            raise TypeError("location is not a string")

        # Get data from api and create objects
        api_datas = self.call_api(fsids, "location", "detail", location_type, limit=limit)

        if location_type == 'property':
            product = [LocationDetailProperty(api_data) for api_data in api_datas]

        elif location_type == 'neighborhood':
            product = [LocationDetailNeighborhood(api_data) for api_data in api_datas]

        elif location_type == 'city':
            product = [LocationDetailCity(api_data) for api_data in api_datas]

        elif location_type == 'zcta':
            product = [LocationDetailZcta(api_data) for api_data in api_datas]

        elif location_type == 'tract':
            product = [LocationDetailTract(api_data) for api_data in api_datas]

        elif location_type == 'county':
            product = [LocationDetailCounty(api_data) for api_data in api_datas]

        elif location_type == 'cd':
            product = [LocationDetailCd(api_data) for api_data in api_datas]

        elif location_type == 'state':
            product = [LocationDetailState(api_data) for api_data in api_datas]

        else:
            raise NotImplementedError

        if csv:
            csv_format.to_csv(product, "location", "detail", location_type, output_dir=output_dir)

        logging.info("Location Detail Data Ready.")

        return product

    def get_summary(self, fsids, location_type, csv=False, limit=100, output_dir=None):
        """Retrieves location summary product data from the First Street Foundation API given a list of FSIDs and
        returns a list of Location Summary objects.

        Args:
            fsids (list/file): A First Street Foundation IDs or a file of First Street Foundation IDs
            location_type (str): The location lookup type
            csv (bool): To output extracted data to a csv or not
            limit (int): max number of connections to make
            output_dir (str): The output directory to save the generated csvs
        Returns:
            A list of Location Summary
        Raises:
            InvalidArgument: The location provided is empty
            TypeError: The location provided is not a string
        """

        if not location_type:
            raise InvalidArgument(location_type)
        elif not isinstance(location_type, str):
            raise TypeError("location is not a string")

        # Get data from api and create objects
        api_datas = self.call_api(fsids, "location", "summary", location_type, limit=limit)

        if location_type == "property":
            product = [LocationSummaryProperty(api_data) for api_data in api_datas]

        else:
            product = [LocationSummaryOther(api_data) for api_data in api_datas]

        if csv:
            csv_format.to_csv(product, "location", "summary", location_type, output_dir=output_dir)

        logging.info("Location Summary Data Ready.")

        return product



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation
import logging


def read_fsid_file(file_name):
    """Reads the given file and pulls a list of FSIDs from the file

    Args:
        file_name (str): A file name
    Returns:
        A list of fsids
    """

    fsids = []

    with open(file_name) as fp:

        count = 1
        for line in fp:

            try:
                fsids.append(int(line.rstrip('\n')))
                count += 1

            except ValueError:
                logging.warning("'{}' from file at line {} is not a valid fsid. This line has been skipped. "
                                "Please check the content of the file if this is unexpected.".
                                format(line.rstrip('\n'), count))

    return fsids



from setuptools import setup, find_packages

with open("README.md", 'r') as f:
    readme = f.read()

with open('requirements.txt') as f:
    requirements = [x.strip() for x in f.readlines()]

with open('extra_test_requires.txt') as f:
    extra = {'testing': [x.strip() for x in f.readlines()]}

setup(
    name='fsf-api-access_python',
    version='0.1.4',
    description='A Python API Access Client for the First Street Foundation API',
    url='https://github.com/FirstStreet/fsf_api_access_python',
    project_urls={
        'First Street Foundation Website': 'https://firststreet.org/'
    },
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Kelvin",
    author_email="kelvin@firststreet.org",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],

    # Package info
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    py_modules=[],
    install_requires=requirements,
    python_requires='>=3.6',
    extras_require=extra
)



"""A Python module for interacting with the First Street Foundation API"""
# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation

# Standard Imports
import logging

# Internal Imports
from firststreet.api.adaptation import Adaptation
from firststreet.api.environmental import Environmental
from firststreet.api.fema import Fema
from firststreet.api.historic import Historic
from firststreet.api.location import Location
from firststreet.api.probability import Probability
from firststreet.errors import MissingAPIKeyError
from firststreet.http_util import Http


class FirstStreet:
    """A FirstStreet allows communication with the First Street Foundation API. This handles constructing and sending
        HTTP requests to the First Street Foundation API, and parses any response received into the appropriate object.

        Attributes:
            api_key (str): A string specifying the API key.
            version (str): The version to call the API with
            log (bool): To log the outputs on info level
        Example:
        ```python
            import os
            import firststreet

            fs = firststreet.FirstStreet(os.environ['FIRSTSTREET_API_KEY'])
            property_summary = fs.data_summary.get_property_by_fsid("450350223646")
        ```
        Raises:
            MissingAPIError: If the API is not provided
    """

    def __init__(self, api_key=None, version=None, log=True):

        if not api_key:
            raise MissingAPIKeyError('Missing API Key.')

        if log:
            logging.basicConfig(level=logging.INFO,
                                format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')

        self.http = Http(api_key, version)
        self.location = Location(self.http)
        self.probability = Probability(self.http)
        self.historic = Historic(self.http)
        self.adaptation = Adaptation(self.http)
        self.environmental = Environmental(self.http)
        self.fema = Fema(self.http)



# Author: Kelvin Lai <kelvin@firststreet.org>
# Copyright: This module is owned by First Street Foundation


class FemaNfip:
    """Creates a FEMA NFIP object given a response

    Args:
        response (JSON): A JSON response received from the API
    """

    def __init__(self, response):
        self.fsid = response.get('fsid')
        self.claimCount = response.get('claimCount')
        self.policyCount = response.get('policyCount')
        self.buildingPaid = response.get('buildingPaid')
        self.contentPaid = response.get('contentPaid')
        self.buildingCoverage = response.get('buildingCoverage')
        self.contentCoverage = response.get('contentCoverage')
        self.iccPaid = response.get('iccPaid')
